
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include "xstd/h/iomanip.h"

#include "xstd/rndDistrs.h"
#include "xstd/StrIdentifier.h"
#include "base/RndPermut.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/Viserv.h"
#include "runtime/PubWorld.h"
#include "runtime/PopModel.h"
#include "runtime/Goal.h"
#include "runtime/LogComment.h"
#include "pgl/RobotSym.h"
#include "pgl/SessionSym.h"
#include "pgl/GoalSym.h"
#include "client/WarmupPlan.h"
#include "client/ServerRep.h"
#include "client/Client.h"
#include "client/CltCfg.h"
#include "client/CltOpts.h"


CltCfg::CltCfg(): theRobot(0), theOriginSel(0), theReqTypeSel(0), theReqMethodSel(0),
	thePopModel(0),
	theBusyPeriod(0), theIdlePeriodDur(0), theProxyCycleCnt(0), theWarmupPlan(0),
	thePublicRatio(-1), theRecurRatio(-1), theEmbedRecurRatio(-1),
	theAbortProb(0),
	theWaitXactLmt(-1), theIcpPort(-1),
	theUsernameCycleCnt(0),
	genUniqUrls(false), didWarmup(false) {
}

CltCfg::~CltCfg() {
	while (theUsernames.count())
		delete theUsernames.dequeue();
	while (theProxies.count())
		delete theProxies.dequeue();
}

void CltCfg::configure(const RobotSym *aRobot) {
	Assert(!theRobot && aRobot);
	theRobot = aRobot;

	theRobot->waitXactLimit(theWaitXactLmt);
	thePeerHttp = theRobot->peerHttp();
	thePeerIcp = theRobot->peerIcp();
	theRobot->icpPort(theIcpPort);

	theRobot->publicRatio(thePublicRatio);
	theRobot->recurRatio(theRecurRatio);
	theRobot->embedRecurRatio(theEmbedRecurRatio);
	theRobot->abortProb(theAbortProb);
	theRobot->uniqueUrls(genUniqUrls);

	if (PopModelSym *pms = theRobot->popModel()) {
		thePopModel = new PopModel;
		thePopModel->configure(pms);
	}

	configureReqTypes();
	configureReqMethods();
	configureOrigins();
	configureProxies();
	configureUsernames();

	if (SessionSym *ss = theRobot->session()) {
		if (GoalSym *bps = ss->busyPeriod()) {
			theBusyPeriod = new Goal;
			theBusyPeriod->configure(bps);
		}
		theIdlePeriodDur = ss->idlePeriodDuration();
		ss->heartbitGap(theSessionHeartbitGap);

		if (!theBusyPeriod != !theIdlePeriodDur) {
			cerr << theRobot->loc() << "both busy and idle periods "
				<< "should be specified (or not specified) for a robot session"
				<< endl << xexit;
		}
	}

	if (!thePopModel && theRecurRatio > 0)
		cerr << theRobot->loc() << "popularity model must be specified for"
			<< " positive recurrence ratio (robot " << theRobot->kind() << ')'
			<< endl << xexit;
}

void CltCfg::configureReqTypes() {
	static StrIdentifier sidf;
	if (!sidf.count()) {
		sidf.add("Basic", Client::rqtBasic);
		sidf.add("IMS", Client::rqtIms);
		sidf.add("Reload", Client::rqtReload);
	}

	theReqTypeSel = theRobot->msgTypes(sidf);
	if (!theReqTypeSel)
		theReqTypeSel = new ConstDistr(new RndGen, Client::rqtBasic); // default
	theReqTypeSel->rndGen(LclRndGen("client_req_types"));
}

void CltCfg::configureReqMethods() {
	static StrIdentifier sidf;
	if (!sidf.count()) {
		sidf.add("GET", Client::rqmGet);
		sidf.add("HEAD", Client::rqmHead);
		sidf.add("POST", Client::rqmPost);
		sidf.add("PUT", Client::rqmPut);
	}

	theReqMethodSel = theRobot->reqMethods(sidf);
	if (!theReqMethodSel)
		theReqMethodSel = new ConstDistr(new RndGen, Client::rqmGet); // default
	theReqMethodSel->rndGen(LclRndGen("client_req_methods"));
}

void CltCfg::configureOrigins() {
	Array<NetAddr*> origNames;

	RndDistr *iad = 0;
	if (!theRobot->origins(origNames, theOriginSel) || !origNames.count()) {
		if (!theRobot->reqInterArrival(iad) || iad)
			cerr << theRobot->loc() << "no origin addresses specified for active robot " 
				<< theRobot->kind() << endl << xexit;
		return;
	}

	Assert(theOriginSel);
	theOriginSel->rndGen(LclRndGen("client_origins"));

	theOrigins.stretch(origNames.count());
	for (int i = 0; i < origNames.count(); ++i) {
		addOrigName(*origNames[i]);
		cAddServers(theOrigins[i]);
	}
}

// converts origin name into viserv idx
void CltCfg::addOrigName(const NetAddr &oname) {
	int nameIdx = -1;
	if (!TheAddrMap->find(oname, nameIdx)) {
		if (oname.isDomainName())
			cerr << here << "visible server name " << oname
				<< " is not found in address maps" << endl << xexit;
		TheAddrMap->add(oname);
	}

	int viserv = -1;
	if (!TheViservs.find(nameIdx, viserv)) {
		TheViservs.add(nameIdx);
		ThePubWorlds.append(new PubWorld(UniqId::Create()));
	}

	theOrigins.append(viserv);
}

// creates new server representatives for vname
void CltCfg::cAddServers(int viserv) {
	const int visName = TheViservs.nameIdxAt(viserv);
	for (AddrMapAddrIter i = TheAddrMap->addrIter(visName); i; ++i) {
		const NetAddr &addr = i.addr();
		int idx = -1;
		if (!TheHostMap->find(addr, idx)) {
			Comment << theRobot->loc() << "Robot cannot find"
				<< " configuration for server@" << addr
				<< " (visible as " << i.name() << ")"
				<< endc << xexit;
		}

		// create representative if needed
		HostCfg *cfg = TheHostMap->at(idx);
		if (!cfg->theServerRep)
			cfg->theServerRep = new ServerRep(addr, idx);
	}
}

void CltCfg::configureProxies() {
	Array<NetAddr*> addrs;
	theRobot->proxies(addrs);
	theProxies.resize(addrs.count());
	while (addrs.count())
		theProxies.enqueue(addrs.pop());

	if (TheCltOpts.theProxyAddr && theProxies.count()) {
		Comment << theRobot->loc() << "--proxy option and "
			<< "Robot.proxies field are mutually exclusive" << endc << xexit;
	}
}

void CltCfg::configureUsernames() {
	Array<String*> names;
	theRobot->usernames(names);
	theUsernames.resize(names.count());
	while (names.count())
		theUsernames.enqueue(names.pop());
}

bool CltCfg::selectUsername(String &newName) {
	if (theUsernames.empty())
		return false;

	// randomize in the beginning of each cycle
	if (theUsernameCycleCnt >= theUsernames.count())
		theUsernameCycleCnt = 0;
	if (!theUsernameCycleCnt) {
		static RndGen rng;
		theUsernames.randomize(rng);
	}
	theUsernameCycleCnt++;
		
	String *name = theUsernames.dequeue();
	theUsernames.enqueue(name);
	newName = *name;
	return true;
}

int CltCfg::selectViserv() {
	if (theWarmupPlan) {
		const int viservIdx = theWarmupPlan->selectViserv();
		if (viservIdx >= 0)
			return viservIdx;
		stopWarmup();
	}

	const int viservIdx = (int)theOriginSel->trial();
	Assert(0 <= viservIdx && viservIdx <= theOrigins.count());
	return theOrigins[viservIdx];
}

bool CltCfg::selectProxy(NetAddr &newAddr) {
	if (TheCltOpts.theProxyAddr) {
		newAddr = TheCltOpts.theProxyAddr;
		return true;
	}

	if (!theProxies.count())
		return false;
	
	// randomize in the beginning of each cycle
	if (theProxyCycleCnt >= theProxies.count())
		theProxyCycleCnt = 0;
	if (!theProxyCycleCnt) {
		static RndGen rng;
		theProxies.randomize(rng);
	}
	theProxyCycleCnt++;
	
	NetAddr *addr = theProxies.dequeue();
	theProxies.enqueue(addr);
	newAddr = *addr;
	return true;
}

// merge this with SrvCfg::selectAbortSize() if more methods can be shared
Size CltCfg::selectAbortSize(Size clen) {
	static RndGen rng1, rng2; // uncorrelated unless theAbortProb is 1
	if (rng1.event(theAbortProb))
		return rng2(0, (int)clen); // OK: abort size of clen is not useful
	return -1;
}

// may be called multiple times
void CltCfg::startWarmup() {
	if (!theWarmupPlan && !didWarmup)
		theWarmupPlan = new WarmupPlan(theOrigins);
}

void CltCfg::stopWarmup() {
	Assert(theWarmupPlan && !didWarmup);
	delete theWarmupPlan;
	theWarmupPlan = 0;
	didWarmup = true;
}

/* CltSharedCfgs */

CltCfg *CltSharedCfgs::getConfig(const RobotSym *rs) {
	for (int i = 0; i < count(); ++i) {
		if (item(i)->theRobot == rs)
			return item(i);
	}
	return addConfig(rs);
}

CltCfg *CltSharedCfgs::addConfig(const RobotSym *rs) {
	CltCfg *cfg = new CltCfg;
	cfg->configure(rs);
	append(cfg);
	return cfg;
}

