
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include "xstd/h/iostream.h"
#include "xstd/h/iomanip.h"
#include <fstream.h>

#include "xstd/Rnd.h"
#include "base/RndPermut.h"
#include "base/BStream.h"
#include "runtime/LogComment.h"
#include "runtime/ErrorMgr.h"

#include "pgl/MimeSym.h"
#include "pgl/ContentSym.h"
#include "csm/EmbedContMdl.h"

#include "csm/RndBodyIter.h"
#include "csm/CdbBodyIter.h"
#include "csm/ContainerBodyIter.h"
#include "csm/InjectIter.h"

#include "csm/ContentDbase.h"
#include "csm/TextDbase.h"
#include "csm/ObjLifeCycle.h"

#include "csm/ContentCfg.h"

#include "runtime/polyErrors.h"


static const String DefaultContentCfgKind = "some-content";


ContentCfg::ContentCfg(int anId):
	theSize(0), theObjLifeCycle(0), theChbRatio(-1), theChecksumRatio(-1),
	theEmbedCont(0), theCdb(0),
	theTdb(0), theInjGap(0), theInfProb(-1),
	theExtSel(0), thePfxSel(0),
	theNewPerOid(0), theId(anId) {
	theObjLifeCycle = new ObjLifeCycle;
}

ContentCfg::~ContentCfg() {
	delete theObjLifeCycle;
}

void ContentCfg::configure(const ContentSym *cfg) {
	Assert(cfg);

	theKind = cfg->kind();
	if (!theKind)
		theKind = DefaultContentCfgKind;

	if (MimeSym *mime = cfg->mime()) {
		theMimeType = mime->mimeType();
		mime->extensions(theExtensions, theExtSel);
		mime->prefixes(thePrefixes, thePfxSel);
	}
	theSize = cfg->size();	
	theObjLifeCycle->configure(cfg->objLifeCycle());
	cfg->cachable(theChbRatio);
	cfg->checksum(theChecksumRatio);

	if (cfg->hasEmbed()) {
		theEmbedCont = new EmbedContMdl;
		theEmbedCont->configure(cfg);
	}

	if (const String &cdbName = cfg->cdb()) {
		ifstream f(cdbName);
		IBStream is;
		is.configure(&f, cdbName);
		theCdb = new ContentDbase;
		theCdb->load(is);
		if (!is)
			Comment << "error: cannot load content database from `"
				<< cdbName << "; " << Error::Last() << endc << xexit;
	}

	if (theEmbedCont && theCdb)
		Comment << "error: content cfg `" << theKind << "': "
			<< "cannot support containers together with content_db yet; "
			<< "do not use may_contain with content_db"
			<< endc << xexit;

	if (const String &tdbName = cfg->injectDb()) {
		ifstream f(tdbName);
		theTdb = new TextDbase;
		theTdb->load(f);
		if (f.bad())
			Comment << "error: cannot load text database from `"
				<< tdbName << "; " << Error::Last() << endc << xexit;
		if (!theTdb->count())
			Comment << "error: text database `"
				<< tdbName << " appears to be empty" << endc << xexit;
	}

	theInjGap = cfg->injectGap();

	// XXX: we should put all inject* fields into one PGL object
	if (cfg->infectProb(theInfProb) != (theTdb != 0) || 
		(theTdb != 0) != (theInjGap != 0)) {
		Comment << "error: content cfg `" << theKind << "': "
			<< "`inject_prob' requires `inject_src' requires `inject_gap'"
			<< " and vice versa" 
			<< endc << xexit;
	}
}

const String &ContentCfg::url_ext(int seed) const {
	return pickStr(theExtensions, theExtSel, seed);
}

const String &ContentCfg::url_pfx(int seed) const {
	return pickStr(thePrefixes, thePfxSel, seed);
}

const String &ContentCfg::pickStr(const Strings &strings, RndDistr *sel, int seed) const {
	static String noStr = 0;
	if (const int count = strings.count()) {
		sel->rndGen()->seed(seed);
		const int idx = (int)sel->trial();
		Assert(0 <= idx && idx < count);
		return *strings[idx];
	}
	return noStr;
}

double ContentCfg::compContPerCall(const ContentCfg *cc) const {
	if (cc->id() == id())
		return 1.0;

	if (theEmbedCont)
		return theEmbedCont->compContPerCall(cc);

	return 0.0;
}

void ContentCfg::noteNewContProb(ContentCfg *cc, double newProb) {
	if (cc->id() == id())
		return;

	if (theEmbedCont)
		theEmbedCont->noteNewContProb(cc, newProb);
}

void ContentCfg::newPerOid(double aNewPerOid) {
	Assert(aNewPerOid > 0);
	theNewPerOid = aNewPerOid;
}

// XXX: iterators should be farmed, but it is hard because they
// come in different types (perhaps somebody else should farm them?)
BodyIter *ContentCfg::getBodyIter(const ObjId &oid) {
	if (theCdb) {
		CdbBodyIter *i = new CdbBodyIter;
		i->oid(oid);
		i->cdb(theCdb);

		if (theTdb) {
			RndGen rng(GlbPermut(oid.hash(), rndInjProb));
			// should we inject this object?
			if (rng.event(theInfProb)) {
				InjectIter *inj = new InjectIter; // XXX: Farm these!
				inj->creator(this);
				inj->textDbase(theTdb);
				inj->gap(theInjGap);
				i->injector(inj);
			}
		}

		return i;
	} else
	if (theEmbedCont) {
		ContainerBodyIter *i = new ContainerBodyIter;
		i->oid(oid);
		i->embedContModel(theEmbedCont);
		return i;
	} else {
		RndBodyIter *i = new RndBodyIter;
		i->oid(oid);
		return i;
	}	
}

void ContentCfg::putBodyIter(BodyIter *i) {
	i->stop();
	delete i;
}

void ContentCfg::putInjector(InjectIter *i) {
	delete i; // XXX: Farm these?
}
