
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include "xstd/h/iostream.h"
#include <fstream.h>
#include "xstd/h/strstream.h"
#include "xstd/h/iomanip.h"

#include "base/StatPhaseRec.h"
#include "base/polyLogCats.h"
#include "base/polyLogTags.h"
#include "base/CmdLine.h"
#include "base/opts.h"

#include "logextractors/LogIter.h"
#include "logextractors/LogCatFilter.h"
#include "logextractors/matchAndPrint.h"


class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,     "help",         "list of options"),
			theOutFileName(this, "out <file>",   "redirect console output", "-"),
			theSideName(this,    "side <clt|srv|all>", "name of `side' to extract"),
			thePhases(this,      "phases <list>","names of phases to extract"),
			theObjects(this,     "objects <list>","names of objects to extract")
			{}

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

		virtual bool validate() const;

	public:
		HelpOpt theHelpOpt;
		StrOpt theOutFileName;
		StrOpt theSideName;
		StrArrOpt thePhases;
		StrArrOpt theObjects;
		Array<String*> theFiles;
};

class PhaseGroup: public StatPhaseRec {
	public:
		PhaseGroup(const String &aName);
};

class MyScanner {
	protected:
		typedef const Array<String*> &Strs;

	public:
		MyScanner(Strs phaseNames, Strs objectNames);
		~MyScanner();

		void filter(LogCatFilter *aFilter) { theFilter = aFilter; }

		bool load(ILog &il, const LogEntryPx &px);

		void report() const;
		
	protected:
		void notePhase(const StatPhaseRec &ph);
		void addCat(int cat);

	protected:
		Strs thePhaseNames;
		Strs theObjectNames;

		Array<PhaseGroup*> theGroups;
		LogCatFilter *theFilter;
};

static MyOpts TheOpts;
static strstream ReportBuf;


/* MyOpt */

bool MyOpts::validate() const {
	if (theSideName && theSideName != "clt" && theSideName != "srv" && theSideName != "all")
		cerr << "side name must be `clt' or `srv' or `all'; got: " << theSideName << endl;
	else
		return OptGrp::validate();
	return false;
}

ostream &MyOpts::printAnonym(ostream &os) const {
	return os << "<log_file_name> ...";
}

bool MyOpts::parseAnonym(const Array<const char *> &opts) {
	for (int i = 0 ; i < opts.count(); ++i)
		theFiles.append(new String(opts[i]));

	// add default fname if none are specified
	if (!theFiles.count())
		theFiles.append(new String("-"));

	return theFiles.count() > 0;
}


/* PhaseGroup */

PhaseGroup::PhaseGroup(const String &aName) {
	theName = aName;
}


/* MyScanner */

MyScanner::MyScanner(Strs phaseNames, Strs objectNames):
	thePhaseNames(phaseNames), theObjectNames(objectNames),
	theFilter(0) {
}

MyScanner::~MyScanner() {
	while (theGroups.count()) delete theGroups.pop();
}

bool MyScanner::load(ILog &il, const LogEntryPx &px) {

	if (!theFilter->passed(px))
		return true;

	switch (px.theTag) {
		case lgContTypeKinds: {
			// should be called only once per log
			ContTypeStat::Load(il);
			break;
		}
		case lgStatPhaseRec: {
			StatPhaseRec ph;
			ph.load(il);
			notePhase(ph);
			break;
		}
	}
	return true;
}

void MyScanner::notePhase(const StatPhaseRec &ph) {
	// check if we need this phase
	bool found = !thePhaseNames.count();
	{for (int i = 0; !found && i < thePhaseNames.count(); ++i)
		found = ph.name()== *thePhaseNames[i];
	}
	if (!found)
		return;

	// check if we saw this phase before
	PhaseGroup *group = 0;
	{for (int i = 0; !group && i < theGroups.count(); ++i) {
		if (theGroups[i]->name() == ph.name())
			group = theGroups[i];
	}}

	if (!group) {
		group = new PhaseGroup(ph.name());
		theGroups.append(group);
	}

	group->merge(ph);
}

void MyScanner::report() const {
	if (!theGroups.count()) {
		cerr << "no matching phases found" << endl;
		exit(-1);
	}

	// concatenate all phases
	PhaseGroup group(0);
	for (int i = 0; i < theGroups.count(); ++i) {
		const PhaseGroup &g = *theGroups[i];
		group.concat(g);
	}
	
	group.print(ReportBuf, "");

	ReportBuf << "log.count:\t " << TheOpts.theFiles.count() << endl;

	ReportBuf << ends;
	if (MatchAndPrint(cout, ReportBuf.str(), theObjectNames) == 0) {
		cerr << "no logged objects matched any of the requested object names" << endl;
		exit(-2);
	}
}


/* local routines */

static
void configureLogs(int prec) {
	if (TheOpts.theOutFileName && TheOpts.theOutFileName != "-")
		cout = *new ofstream(TheOpts.theOutFileName);

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);

	configureStream(ReportBuf, prec);
}

static
void configure() {
	configureLogs(2);
}

int main(int argc, char *argv[]) {

	CmdLine cmd;
	cmd.configure(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();

	int lc = lgcAll;
	if (TheOpts.theSideName == "clt")
		lc = lgcCltSide;
	else
	if (TheOpts.theSideName == "srv")
		lc = lgcSrvSide;

	LogCatFilter *filter = new LogCatFilter;
	filter->logCat(lc);

	MyScanner scanner(TheOpts.thePhases.val(), TheOpts.theObjects.val());
	scanner.filter(filter);

	for (int i = 0; i < TheOpts.theFiles.count(); ++i) {
		const String &fname = *TheOpts.theFiles[i];
		ILog log;
		if (fname == "-")
			log.stream("stdin", &cin);
		else
			log.stream(fname, (istream*)new ifstream(fname, ios::binary|ios::in));

		for (LogIter li(&log); li; ++li)
			scanner.load(log, *li);
	}

	scanner.report();
	delete filter;

	return 0;
}
