
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "pgl/pgl.h"

#include "pgl/PglCtx.h"
#include "pgl/PglRec.h"
#include "pgl/PglRecSym.h"

#include "pgl/PglBoolSym.h"
#include "pgl/PglIntSym.h"
#include "pgl/PglNumSym.h"
#include "pgl/PglSizeSym.h"
#include "pgl/PglRateSym.h"
#include "pgl/PglBwidthSym.h"
#include "pgl/PglDistrSym.h"
#include "pgl/PglStringSym.h"
#include "pgl/PglNetAddrSym.h"
#include "pgl/PglTimeSym.h"
#include "pgl/PglArraySym.h"


RecSym::~RecSym() {
	delete theRec;
}

bool RecSym::isA(const String &type) const {
	return ExpressionSym::isA(type) || type == "Rec";
}

SynSym *RecSym::dupe(const String &type) const {
	Assert(theRec);

	if (isA(type))
		return new RecSym(this->type(), theRec->clone());
	else
		return ExpressionSym::dupe(type);
}

bool RecSym::empty() const {
	const int symCount = theRec->count();
	for (int i = 0; i < symCount; ++i) {
		if (const SynSym *ss = theRec->valAt(i)->sym()) {
			// empty() is recursive for items that are records
			if (ss->isA("Rec") && ((const RecSym&)ss->cast("Rec")).empty())
				continue;
			return false;
		}
	}
	return true;
}

ostream &RecSym::print(ostream &os, const String &pfx) const {
	os << '{' << endl;
	if (theRec)
		theRec->report(os, pfx+"    ");
	os << pfx << '}';
	return os;
}

bool RecSym::getBool(const String &name, bool &set) const {
	SynSymTblItem *bi = 0;
	Assert(theRec);
	Assert(theRec->find(name, bi));
	if (bi->sym())
		set = ((const BoolSym&)bi->sym()->cast("bool")).val();
	return bi->sym() != 0;
}

bool RecSym::getInt(const String &name, int &n) const {
	SynSymTblItem *ii = 0;
	Assert(theRec);
	Assert(theRec->find(name, ii));
	if (ii->sym())
		n = ((const IntSym&)ii->sym()->cast("int")).val();
	return ii->sym() != 0;
}

bool RecSym::getDouble(const String &name, double &n) const {
	SynSymTblItem *di = 0;
	Assert(theRec);
	Assert(theRec->find(name, di));
	if (di->sym())
		n = ((const NumSym&)di->sym()->cast("float")).val();
	return di->sym() != 0;
}

Time RecSym::getTime(const String &name) const {
	SynSymTblItem *ti = 0;
	Assert(theRec);
	Assert(theRec->find(name, ti));
	return ti->sym() ?
		((const TimeSym&)ti->sym()->cast("time")).val() : Time();
}

RndDistr *RecSym::getDistr(const String &name) const {
	const String typeName = "Distr";
	SynSymTblItem *di = 0;
	Assert(theRec);
	Assert(theRec->find(name, di));
	return di->sym() ?
		((const DistrSym&)di->sym()->cast(typeName)).val() : 0;
}

String RecSym::getString(const String &name) const {
	SynSymTblItem *ni = 0;
	Assert(theRec->find(name, ni));
	return ni->sym() ?
		((const StringSym&)ni->sym()->cast("string")).val() : String(0);
}

NetAddr RecSym::getNetAddr(const String &name) const {
	SynSymTblItem *ni = 0;
	Assert(theRec->find(name, ni));
	return ni->sym() ?
		((const NetAddrSym&)ni->sym()->cast("addr")).val() : NetAddr();
}

bool RecSym::getSize(const String &name, BigSize &sz) const {
	SynSymTblItem *ni = 0;
	Assert(theRec->find(name, ni));
	if (!ni->sym())
		return false;
	sz = ((const SizeSym&)ni->sym()->cast("size")).val();
	return true;
}

bool RecSym::getRate(const String &name, double &rate) const {
	SynSymTblItem *ri = 0;
	Assert(theRec->find(name, ri));
	if (ri->sym())
		rate = ((const RateSym&)ri->sym()->cast("rate")).val();
	return ri->sym() != 0;
}

bool RecSym::getBwidth(const String &name, Bwidth &bw) const {
	SynSymTblItem *bi = 0;
	Assert(theRec->find(name, bi));
	if (!bi->sym())
		return false;
	bw = ((const BwidthSym&)bi->sym()->cast("bwidth")).val();
	return true;
}

bool RecSym::getNetAddrs(const String &name, Array<NetAddr*> &addrs) const {
	SynSymTblItem *si = 0;
	Assert(theRec->find(name, si));
	if (!si->sym())
		return false;

	ArraySym &ss = (ArraySym&)si->sym()->cast("Array");
	for (int i = 0; i < ss.count(); ++i) {
		const NetAddr &addr = ((const NetAddrSym&)ss[i]->cast("addr")).val();
		addrs.append(new NetAddr(addr));
	}

	return true;
}

ArraySym *RecSym::getArraySym(const String &name) const {
	SynSymTblItem *arrsi = 0;
	Assert(theRec);
	Assert(theRec->find(name, arrsi));
	if (!arrsi->sym())
		return 0;

	ArraySym &arr = (ArraySym&)arrsi->sym()->cast("Array");
	return arr.count() > 0 ? &arr : 0;
}

bool RecSym::getStrings(const String &name, Array<String*> &strs) const {
	if (ArraySym *as = getArraySym(name)) {
		for (int i = 0; i < as->count(); ++i) {
			const String &str = ((const StringSym&)as->item(i)->cast("string")).val();
			strs.append(new String(str));
		}
		return true;
	}
	return false;
}

RecSym *RecSym::getRecSym(const String &name, bool emptyOk) const {
	SynSymTblItem *rsi = 0;
	Assert(theRec);
	Assert(theRec->find(name, rsi));
	if (!rsi->sym())
		return 0;

	RecSym &rs = (RecSym&)rsi->sym()->cast("Rec");
	return rs.empty() && !emptyOk ? 0 : &rs;
}
