
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "pgl/pgl.h"

#include <ctype.h>

#include "xstd/Dns.h"
#include "xstd/Time.h"
#include "xstd/BigSize.h"
#include "xstd/NetAddr.h"
#include "xstd/rndDistrs.h"
#include "xstd/gadgets.h"
#include "xstd/Assert.h"
#include "pgl/pglStrIs.h"


//static String dname2ip(const String &dname);


bool pglIsRatio(const String &str, double &r) {
	if (str) {
		double h;
		const char *ptr = 0;
		if (isNum(str, h, &ptr)) {
			if (!*ptr) { // nothing at the end
				r = h;
				return true;
			} else
			if (!strcmp(ptr, "%") || !strcmp(ptr, "p")) { // percent
				r = h / 100;
				return true;
			}
		}
	}
	return false;
}

bool pglIsRelTime(const String &str, Time &t) {
	// note: things are set up so that "sec" stands for "1sec" 
	//       (convenient for rates)
	const char *ptr = str;
	int i = 1;
	double d;

	// try to preserve integer precision if possible
	const bool isi = !isDbl(str, d, &ptr);

	if (isi)
		isInt(str, i, &ptr);

	if (!*ptr && isi && i == 0)
		t = Time(0,0);
	else
	if (!strcmp(ptr, "msec") || !strcmp(ptr, "ms"))
		t = isi ? Time::Msec(i) : Time::Secd(d/1e3);
	else
	if (!strcmp(ptr, "sec") || !strcmp(ptr, "s"))
		t = isi ? Time::Sec(i) : Time::Secd(d);
	else
	if (!strcmp(ptr, "min"))
		t = isi ? Time::Sec(i*60) : Time::Mind(d);
	else
	if (!strcmp(ptr, "hour") || !strcmp(ptr, "hr"))
		t = isi ? Time::Sec(i*60*60) : Time::Hourd(d);
	else
	if (!strcmp(ptr, "day"))
		t = isi ? Time::Sec(i*24*60*60) : Time::Dayd(d);
	else
	if (!strcmp(ptr, "year"))
		t = isi ? Time::Sec(i*365*24*60*60) : Time::Yeard(d);
	else
		return false;

	return true;
}

bool pglIsAbsTime(const String &str, Time &t) {
	struct tm h;
	memset(&h, 0, sizeof(h));

	const char *eos = str + str.len();

	// check for day: YYYY/MM/DD
	const char *p = str.chr('/');
	if (!p || (p-=4) < str.str() || p+9 >= eos)
		return false;
	
	h.tm_year = xatoi(p, -1) - 1900;
	h.tm_mon = xatoi(p+=5, -1) - 1;
	h.tm_mday = xatoi(p+=3, -1);

	if (h.tm_year <= 0 || h.tm_mon < 0 || h.tm_mday <= 0)
		return false;

	// check for time: HH:MM:SS
	p = str.chr(':');
	if (p) {
		if ((p-=2) < str.str() || p+4 >= eos)
			return false;

		h.tm_hour = xatoi(p, -1);
		h.tm_min = xatoi(p+=3, -1);
		if ((p+=2) < eos && *p == ':')
			h.tm_sec = xatoi(++p, -1);

		if (h.tm_hour < 0 || h.tm_min < 0 || h.tm_sec < 0)
			return false;
	}

	Time tt(h);
	if (tt >= 0) {
		t = tt;
		return true;
	}

	return false;
}

bool pglIsSize(const String &str, BigSize &s) {
	// note: things are set up so that "KB" stands for "1KB" 
	//       (convenient for rates)
	const char *ptr = str;
	int i = 1;
	double d = 1;
	bool isi = false;

	if (isDbl(str, d, &ptr))
		i = (int)rint(d);
	else
	if ((isi = isInt(str, i, &ptr)))
		d = i;

	if (isi && !*ptr && i == 0)
		s = (0);
	else
	if (!strcmp(ptr, "Byte"))
		s = (i);
	else
	if (!strcmp(ptr, "Kb") || !strcmp(ptr, "Kbit"))
		s = ((int)(1024*d/8));
	else
	if (!strcmp(ptr, "KB") || !strcmp(ptr, "KByte"))
		s = ((int)(1024*d));
	else
	if (!strcmp(ptr, "Mb") || !strcmp(ptr, "Mbit"))
		s = BigSize::MBd(d/8);
	else
	if (!strcmp(ptr, "MB") || !strcmp(ptr, "MByte"))
		s = isi ? BigSize::MB((int)d) : BigSize::MBd(d);
	else
	if (!strcmp(ptr, "Gb") || !strcmp(ptr, "Gbit"))
		s = BigSize::MBd(1024*d/8);
	else
	if (!strcmp(ptr, "GB") || !strcmp(ptr, "GByte"))
		s = BigSize::MBd(1024*d);
	else
		return false;

	return true;
}

bool pglSplitAddr(const String &str, String &ifname, String &addr, int &port, int &subnet) {
	String val = str;

	String parsedIfname;
	int parsedPort = -1;
	int parsedSubnet = -1;

	if (!pglSplitAddr(str, parsedIfname, val, parsedSubnet))
		return false;

	if (const char *p = val.chr(':')) { // port
		if (!isInt(p+1, parsedPort))
			return false;
		val = val(0, p-val);
	}

	// what is left must be a host name (IP or dname) or name range
	if (!val.len())
		return false;

	ifname = parsedIfname;
	addr = val;
	port = parsedPort;
	subnet = parsedSubnet;
	return true;
}

bool pglSplitAddr(const String &str, String &ifname, String &addr, int &subnet) {
	String val = str;

	String parsedIfname;
	int parsedSubnet = -1;

	if (const char *p = val.str("::")) { // ifname
		parsedIfname = val(0, p-val);
		val = p+2;
	}

	if (const char *p = val.chr('/')) { // subnet
		if (!isInt(p+1, parsedSubnet))
			return false;
		if (!(1 <= parsedSubnet && parsedSubnet <= 32))
			return false;
		val = val(0, p-val);
	}

	// what is left must be a host name (IP or dname) or name range
	// possibly with a port or port range
	if (!val.len())
		return false;

	// check that all characters are valid
	for (const char *c = val.str(); *c; ++c) {
		if (!isalpha(*c) && !isdigit(*c) && 
			*c != '-' && *c != '[' && *c != ']' && 
			*c != ':' && *c != '.')
			return false;
	}
		

	ifname = parsedIfname;
	addr = val;
	subnet = parsedSubnet;
	return true;
}

bool pglIsNetAddr(const String &str, String &ifname, String &host, int &port, int &subnet) {
	if (str.chr('-'))
		return false;

	if (!pglSplitAddr(str, ifname, host, port, subnet))
		return false;

//	if (isalpha(host[0])) { // dname
//		host = dname2ip(host);
//	}

	return true;
}

NetAddr pglIsNetAddr(const String &str) {
	String ifname;
	String host;
	int port = -1;
	int subnet = -1;

	if (!pglSplitAddr(str, ifname, host, port, subnet))
		return NetAddr();

//	if (isalpha(host[0])) { // dname
//		host = dname2ip(host);
//	}

	return NetAddr(host, port);
}

#if 0
static
String dname2ip(const String &dname) {
	if (hostent *he = Dns::GetHostByName(dname)) {
		const char *str0 = he->h_addr_list[0];
		if (str0 && he->h_addr_list[1]) {
			cerr << here << "warning: domain name `" << dname 
				<< "' resolves to multiple addresses, using the first one" 
				<< endl;
		}
		struct in_addr addr;
		memcpy(&addr, str0, sizeof(addr));
		const String ip = inet_ntoa(addr);
		Must(ip && ip.len());
		return ip;
	}

	// XXX: convert to Error interface
	cerr << here << "error: cannot resolve `" << dname << "' domain name: " << hstrerror(h_errno) << endl << xexit;
	return 0;
}
#endif
