
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#ifndef XSTD_FILE_SCANNER_H
#define XSTD_FILE_SCANNER_H

#include "xstd/h/sys/types.h"
#include "xstd/h/os_std.h"

#include "xstd/Array.h"
#include "xstd/Time.h"

// file scanner is a common interface to select(2) and poll(2)

typedef enum { dirNone = -1, dirRead = 0, dirWrite = 1 } IODir;

class FileScanUser {
	public:
		virtual ~FileScanUser();

		// default implementation will just abort()
		virtual void noteReadReady(int fd);
		virtual void noteWriteReady(int fd);
		virtual void noteTimeout(int fd, Time tout);
};

// reservation receipt
class FileScanReserv {
	public:
		FileScanReserv();
		FileScanReserv(int aFD, IODir aDir);

		operator bool() const { return reserved(); }
		bool reserved() const { return theFD >= 0; }
		bool ready() const { return isReady; }
		int fd() const { return theFD; }
		IODir dir() const { return theDir; }

		void ready(bool be) { isReady = be; }
		void clear() { theFD = -1; }

	protected:
		int theFD;
		IODir theDir;
		bool isReady;
};

// Ticker is notified about various stages of a scan process
class FileScanTicker {
	public:
		virtual ~FileScanTicker() {}

		virtual void begCycle(int readyCount) = 0;
		virtual void endCycle() = 0;
		virtual bool tick() = 0;
};

typedef enum { fsupMin, fsupBestEffort, fsupAsap, fsupDefault } FileScanUserPriority;

// an internal registration record for a user to monitor I/O activity
class FileScanUserReg {
	public:
		typedef FileScanUserPriority Priority;

	public:
		FileScanUserReg() {}

		void reset();

		void set(FileScanUser *aUser); // updates start time and res counter
		void clear(); // decrements reservation counter
		void timeout(Time aTout);
		void priority(Priority aPriority);

		FileScanUser *user() { return theUser; }
		Priority priority() const { return thePriority; }
		Time waitTime() const;

		bool timedout() const;
		bool needsCheck() const;

		void noteCheck(bool didIO);

	protected:
		FileScanUser *theUser;
		Time theTimeout;         // how long a user is willing to wait for an io
		Time theStart;           // registration or last io
		int theResCount;         // how many reservations (e.g., read + write)
		Priority thePriority;
};

// the "scanner" itself
class FileScanner {
	friend class FileScanUserReg;

	public:
		typedef FileScanUser User;
		typedef FileScanTicker Ticker;
		typedef FileScanUserPriority Priority;

	protected:
		static Priority TheMinPriority;  // min priority to be scan()ned

	public:
		FileScanner();
		virtual ~FileScanner();

		// scanner's name (e.g., system call for syscall wrappers)
		virtual const char *name() const = 0;

		virtual void configure(int fdLimit);
		void ticker(Ticker *aTicker);

		bool idle() const { return theResCount <= 0; }

		virtual FileScanReserv setFD(int fd, IODir dir, User *p);
		virtual void clearFD(int fd, IODir dir);
		void clearRes(FileScanReserv &res) { clearFD(res.fd(), res.dir()); res.clear(); }

		void setTimeout(int fd, Time tout);
		void clearTimeout(int fd);

		void setPriority(int fd, Priority p);
		Priority getPriority(int fd) const;

		// returns a positive number if some files where ready; -1 on error
		// calls applicable User methods for ready files
		// null timeout means wait until error or at least one file is ready
		int scan(Priority minP, Time *timeout = 0);
		int scan(Time *tout = 0) { return scan(fsupMin, tout); }

	protected:
		// returns "hot count" (see scan code)
		virtual int sweep(Time *timeout = 0) = 0;

		// must set fd if the indexed user is ready
		virtual User *readyUser(int idx, IODir dir, int &fd) = 0;

		void checkTimeouts();

	protected:
		Array<FileScanUserReg> theRegs; // user registrations

		int theMaxFD;
		int theResCount;
		Ticker *theTicker;
};

// we must add an iterator to make the interface complete

#endif
